from typing import Dict, List
from dataclasses import dataclass
from queenbee.io.inputs.function import (
    FunctionStringInput, FunctionIntegerInput, FunctionNumberInput,
    FunctionBooleanInput, FunctionFolderInput, FunctionFileInput,
    FunctionPathInput, FunctionJSONObjectInput, FunctionArrayInput
)
from queenbee.base.basemodel import BaseModel, Field
from queenbee.io.common import ItemType

__all__ = ('Inputs', )

_inputs_mapper = {
    'StringInput': FunctionStringInput,
    'IntegerInput': FunctionIntegerInput,
    'NumberInput': FunctionNumberInput,
    'BooleanInput': FunctionBooleanInput,
    'FolderInput': FunctionFolderInput,
    'FileInput': FunctionFileInput,
    'PathInput': FunctionPathInput,
    'DictInput': FunctionJSONObjectInput,
    'ListInput': FunctionArrayInput
}


class _InputBase(BaseModel):

    @property
    def __decorator__(self) -> str:
        """Queenbee decorator for inputs."""
        return 'input'

    def to_queenbee(self, name):
        """Convert this input to a Queenbee input."""
        func = _inputs_mapper[self.__class__.__name__]
        data = {
            'required': True if self.default is None else False,
            'name': name.replace('_', '-'),
            'default': self.default,
            'description': self.description,
            'annotations': self.annotations,
            'spec': self.spec
        }

        if hasattr(self, 'path'):
            data['path'] = self.path

        if hasattr(self, 'extensions'):
            data['extensions'] = self.extensions

        return func.parse_obj(data)


class StringInput(_InputBase):
    """ A Function string input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.

    """
    annotations: Dict = None
    description: str = None
    default: str = None
    spec: Dict = None


class IntegerInput(StringInput):
    """ A Function integer input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.

    """
    default: int = None


class NumberInput(StringInput):
    """ A Function number input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.

    """
    default: float = None


class BooleanInput(StringInput):
    """ A Function boolean input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.

    """
    default: bool = None


class DictInput(StringInput):
    """ A Function dictionary input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.

    """
    default: Dict = None


class ListInput(StringInput):
    """ A Function list input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        items_type: 'Type of items in list. All the items in an array must be from '
        'the same type.'
        spec: A JSONSchema specification to validate input values.

    """
    default: List = None

    items_type: ItemType = Field(
        ItemType.String,
        description='Type of items in an array. All the items in an array must be from '
        'the same type.'
    )


class FolderInput(StringInput):
    """ A Function folder input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        path: Path to source folder.
        spec: A JSONSchema specification to validate input values.

    """
    path: str


class FileInput(FolderInput):
    """ A Function file input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        extensions: An optional list of valid extensions for input file.
        path: Path to source folder.
        spec: A JSONSchema specification to validate input values.

    """
    extensions: List[str] = None


class PathInput(FileInput):
    """ A Function file input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        extensions: An optional list of valid extensions for input file.
        path: Path to source folder.
        spec: A JSONSchema specification to validate input values.

    """
    ...


@dataclass
class Inputs:
    """Function inputs enumeration."""
    str = StringInput
    int = IntegerInput
    float = NumberInput
    bool = BooleanInput
    file = FileInput
    folder = FolderInput
    path = PathInput
    dict = DictInput
    list = ListInput



from dataclasses import dataclass
from typing import NamedTuple
import importlib
from collections import namedtuple


def camel_to_snake(name: str) -> str:
    """Change name from CamelCase to snake-case."""
    return name[0].lower() + \
        ''.join(['-' + x.lower() if x.isupper() else x for x in name][1:])


def import_module(name: str):
    """Import a module by name.

    This function works for both namespace and non-name space Python modules.
    """
    package_name = name.replace('-', '_')
    err_msg = \
        f'No module named \'{package_name}\'. Did you forget to install the module?\n' \
        'You can use `pip install` command to install the package from a local ' \
        'repository or from PyPI.'
    try:
        module = importlib.import_module(package_name)
    except ModuleNotFoundError:
        # for pollination modules split and try again pollination-honeybee-radiance
        # is namedspaced as pollination.honeybee_radiance
        package_name_segments = package_name.split('_')
        if len(package_name_segments) == 1:
            raise ModuleNotFoundError(err_msg)
        _namespace = package_name_segments[0]
        _name = '_'.join(package_name_segments[1:])
        try:
            namespace = __import__(f'{_namespace}.{_name}')
            module = getattr(namespace, _name)
        except ModuleNotFoundError:
            raise ModuleNotFoundError(err_msg)

    return module


@dataclass
class _BaseClass:
    """Base class for Pollination dsl Function and DAG.

    Do not use this class directly.
    """
    _cached_queenbee = None
    _cached_outputs = None
    _cached_package = None
    _cached_inputs = None

    @property
    def queenbee(self):
        raise NotImplementedError

    @property
    def _outputs(self) -> NamedTuple:
        raise NotImplementedError

    @property
    def _inputs(self) -> NamedTuple:
        """Return inputs as a simple object with dot notation.

        Use this property to access the inputs when creating a DAG.

        The name starts with a _ not to conflict with a possible member of the class
        with the name inputs.
        """
        if self._cached_inputs:
            return self._cached_inputs
        cls_name = camel_to_snake(self.__class__.__name__)
        mapper = {
            inp.name.replace('-', '_'): {
                'name': inp.name.replace('-', '_'),
                'parent': cls_name,
                'value': inp
            } for inp in self.queenbee.inputs
        }

        inputs = namedtuple('Inputs', list(mapper.keys()))
        self._cached_inputs = inputs(*list(mapper.values()))

        return self._cached_inputs

    @property
    def _package(self) -> dict:
        """Queenbee package information.

        This information will only be available if the function is part of a Python
        package.
        """
        if self._cached_package:
            return self._cached_package

        module = importlib.import_module(self._python_package)
        assert hasattr(module, '__pollination__'), \
            'Failed to find __pollination__ info in __init__.py'
        self._cached_package = getattr(module, '__pollination__')
        return self._cached_package

    @property
    def _python_package(self) -> str:
        """Python package information for this function.

        This information will only be available if the function is part of a Python
        package.
        """
        return self.__module__.split('.')[0]



from pollination_dsl.alias.outputs import OutputAliasTypes
from typing import Any, Dict, List
from dataclasses import dataclass

from queenbee.io.outputs.dag import (
    DAGGenericOutput, DAGStringOutput, DAGIntegerOutput, DAGNumberOutput,
    DAGBooleanOutput, DAGFolderOutput, DAGFileOutput, DAGPathOutput,
    DAGJSONObjectOutput, DAGArrayOutput
)
from queenbee.base.basemodel import BaseModel
from queenbee.io.common import ItemType
from queenbee.base.parser import parse_double_quotes_vars

from pydantic import validator, Field


__all__ = ('Outputs', )

_outputs_mapper = {
    'GenericOutput': DAGGenericOutput,
    'StringOutput': DAGStringOutput,
    'IntegerOutput': DAGIntegerOutput,
    'NumberOutput': DAGNumberOutput,
    'BooleanOutput': DAGBooleanOutput,
    'FolderOutput': DAGFolderOutput,
    'FileOutput': DAGFileOutput,
    'PathOutput': DAGPathOutput,
    'DictOutput': DAGJSONObjectOutput,
    'ListOutput': DAGArrayOutput
}


def _get_from(value, reference):
    try:
        # only TaskReference has parent
        parent = value['parent']
    except TypeError:
        # value
        return {'type': reference, 'path': value}
    else:
        # Task reference values and names are nested
        variable = value['name']
        value = value['value']
        return {'type': reference, 'name': parent, 'variable': variable}


class _OutputBase(BaseModel):

    source: Any  # this field will be translated to from_
    annotations: Dict = None
    description: str = None
    alias: List[OutputAliasTypes] = None

    @validator('source')
    def change_self_to_inputs(cls, v):
        refs = parse_double_quotes_vars(v)
        for ref in refs:
            v = v.replace(
                ref, ref.replace('self.', 'inputs.').replace('_', '-')
            )
        return v

    @validator('alias', always=True)
    def empty_list_alias(cls, v):
        return v if v is not None else []

    @property
    def __decorator__(self) -> str:
        """Queenbee decorator for outputs."""
        return 'output'

    def to_queenbee(self, name):
        """Convert this output to a Queenbee DAG output."""
        func = _outputs_mapper[self.__class__.__name__]
        data = {
            'name': name.replace('_', '-'),
            'from': _get_from(self.source, self.reference_type),
            'description': self.description,
            'annotations': self.annotations,
            'alias': [al.to_queenbee().dict() for al in self.alias]
        }

        if hasattr(self, 'items_type'):
            data['items_type'] = self.items_type

        return func.parse_obj(data)

    @property
    def is_artifact(self):
        return False

    @property
    def reference_type(self):
        return 'TaskReference'


class GenericOutput(_OutputBase):
    """ A DAG generic output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """
    ...


class StringOutput(GenericOutput):
    """ A DAG string output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """

    source: Any  # this field will be translated to from_
    annotations: Dict = None
    description: str = None

    @validator('source')
    def change_self_to_inputs(cls, v):
        refs = parse_double_quotes_vars(v)
        for ref in refs:
            v = v.replace(
                ref, ref.replace('self.', 'inputs.').replace('_', '-')
            )
        return v


class IntegerOutput(StringOutput):
    """ A DAG integer output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """
    ...


class NumberOutput(StringOutput):
    """ A DAG number output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """
    ...


class BooleanOutput(StringOutput):
    ...


class DictOutput(StringOutput):
    """ A DAG dictionary output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """
    ...


class ListOutput(StringOutput):
    """ A DAG list output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """
    items_type: ItemType = Field(
        ItemType.String,
        description='Type of items in this array. All the items in an array must be '
        'from the same type.'
    )


class FolderOutput(StringOutput):
    """ A DAG folder output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """

    @property
    def is_artifact(self):
        return True

    @property
    def reference_type(self):
        return 'FolderReference'


class FileOutput(FolderOutput):
    """ A DAG file output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """

    @property
    def reference_type(self):
        return 'FileReference'


class PathOutput(FolderOutput):
    """ A DAG path output. A path can be a file or a folder.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        source: Source for this output. A source is usually from one of the template
            outputs but it can also be declared as a relative path.
        alias: A list of aliases for this output in different platforms.

    """
    ...


@dataclass
class Outputs:
    """DAG outputs enumeration."""
    any = GenericOutput
    str = StringOutput
    int = IntegerOutput
    float = NumberOutput
    bool = BooleanOutput
    file = FileOutput
    folder = FolderOutput
    path = PathOutput
    dict = DictOutput
    list = ListOutput



import importlib
import pkgutil
import pkg_resources
import pathlib
import importlib_metadata
import re

from setuptools.command.develop import develop
from setuptools.command.install import install
from typing import Dict, List, Union

from queenbee.plugin.plugin import Plugin, PluginConfig, MetaData
from queenbee.recipe.recipe import Recipe, BakedRecipe, Dependency, DependencyKind
from queenbee.repository.package import PackageVersion
from queenbee.repository.index import RepositoryIndex
from queenbee.config import Config, RepositoryReference

from .function import Function
from .common import import_module


def _init_repo() -> pathlib.Path:
    """Initiate a local Queenbee repository.

    This function is used by package function to start a local Queenbee repository
    if it doesn't exist. If the repository has already been created it will return
    the path to the repository.
    """

    path = pathlib.Path.home()/'.queenbee'/'pollination-dsl'
    path.mkdir(exist_ok=True)
    index_file = path/'index.json'
    if index_file.exists():
        return path

    plugins_folder = path/'plugins'
    recipes_folder = path/'recipes'
    plugins_folder.mkdir(exist_ok=True)
    recipes_folder.mkdir(exist_ok=True)
    index = RepositoryIndex.from_folder(path.as_posix())

    index.to_json(index_file.as_posix(), indent=2)
    return path


class PostInstall(install):
    """A class to extend `pip install` run method.

    By adding this class to setup.py this package will be added to pollination-dsl local
    repository which makes it accessible to other queenbee packages as a dependency.

    See here for an example: https://github.com/pollination/honeybee-radiance-pollination/blob/0b5590f691427f256beb77b37bd43f545106eaf1/setup.py#L3-L14
    """

    def run(self):
        install.run(self)
        # add queenbee package to pollination-dsl repository
        package_name = self.config_vars['dist_name']
        package(package_name)


class PostDevelop(develop):
    """A class to extend `pip install -e` run method.

    By adding this class to setup.py this package will be added to pollination-dsl local
    repository which makes it accessible to other queenbee packages as a dependency.

    See here for an example: https://github.com/pollination/honeybee-radiance-pollination/blob/0b5590f691427f256beb77b37bd43f545106eaf1/setup.py#L3-L14
    """

    def run(self):
        develop.run(self)
        package_name = self.config_vars['dist_name']
        package(package_name)


def get_requirement_version(package_name, dependency_name):
    """Get assigned version to a dependency in package requirements."""
    req = pkg_resources.get_distribution(
        package_name.replace('pollination.', 'pollination_')
    ).get_metadata('requires.txt')
    requirements = {}
    for package in pkg_resources.parse_requirements(req):
        version = \
            str(package.specifier).replace('=', '').replace('>', '').replace('<', '')
        requirements[package.project_name] = version

    assert dependency_name.replace('_', '-') in requirements, \
        f'{dependency_name} is not a requirement for {package_name}.'

    return requirements[dependency_name.replace('_', '-')]


def _get_package_readme(package_name: str) -> str:
    package_data = importlib_metadata.metadata(package_name.replace('-', '_'))
    long_description = package_data.get_payload()
    if not long_description.strip():
        content = package_data.get('Description')
        long_description = []
        for line in content.split('\n'):
            long_description.append(re.sub('^        ', '', line))
        long_description = '\n'.join(long_description)
    return long_description


def _get_package_owner(package_name: str) -> str:
    """Author field is used for package owner."""
    package_data = importlib_metadata.metadata(package_name.replace('-', '_'))
    owner = package_data.get('Author')
    assert owner, \
        'You must set the author of the package in setup.py to Pollination account owner'
    # ensure there is only one author
    owner = owner.strip()
    assert len(owner.split(',')) == 1, \
        'A Pollination package can only have one author. Use maintainer field for ' \
        'providing multiple maintainers.'

    return owner


def _get_package_license(package_data: Dict) -> Dict:
    # try to get license
    license_info = package_data.get('License')
    if not license_info:
        license, link = None, None
    elif license_info and ',' in license_info:
        license, link = [info.strip() for info in license_info.split(',')]
    else:
        license, link = license_info.strip(), None

    return {'name': license, 'url': link}


def _get_package_keywords(package_data: Dict) -> List:
    keywords = package_data.get('Keyword')
    if keywords:
        keywords = [key.strip() for key in keywords.split(',')]
    return keywords


def _get_package_icon(package_data: Dict) -> str:
    urls = package_data.get_all('Project-URL')
    for url in urls:
        key, value = url.split(',')
        if key == 'icon':
            return value.strip()


def _get_package_maintainers(package_data: Dict) -> List[Dict]:
    package_maintainers = []
    maintainer = package_data.get('Maintainer')
    maintainer_email = package_data.get('Maintainer-email')

    if maintainer:
        maintainers = [m.strip() for m in maintainer.split(',')]
        if maintainer_email:
            emails = [m.strip() for m in maintainer_email.split(',')]
            for name, email in zip(maintainers, emails):
                package_maintainers.append({'name': name, 'email': email})
        else:
            for name in maintainers:
                package_maintainers.append({'name': name, 'email': None})
    return package_maintainers


def _get_package_version(package_data: Dict) -> str:
    """Get package version.

    This function returns the non-development version for a development version.
    It removes the .dev part and return x.y.z-1 version if it is a dev version.
    """
    version = package_data.get('Version')
    xyz = version.split('.')
    if len(xyz) <= 3:
        return version
    # clean up the developer version
    x, y, z = xyz[:3]
    return f'{x}.{y}.{int(z) - 1}'


def _get_package_data(package_name: str) -> Dict:

    package_data = importlib_metadata.metadata(package_name)

    data = {
        'name': package_data.get('Name').replace('pollination-', ''),
        'description': package_data.get('Summary'),
        'home': package_data.get('Home-page'),
        'tag': _get_package_version(package_data),
        'keywords': _get_package_keywords(package_data),
        'maintainers': _get_package_maintainers(package_data),
        'license': _get_package_license(package_data),
        'icon': _get_package_icon(package_data)
    }

    return data


def _get_meta_data(module, package_type: str) -> MetaData:
    qb_info = module.__pollination__
    package_data = _get_package_data(module.__name__)

    if package_type == 'plugin':
        qb_info.pop('config')
    else:
        # recipe
        qb_info.pop('entry_point')

    for k, v in package_data.items():
        if v is None:
            continue
        if k == 'name' and k in qb_info:
            # only use package name if name is not provided
            continue
        qb_info[k] = v

    metadata = MetaData.parse_obj(qb_info)

    return metadata


def _load_plugin(module) -> Plugin:
    """Load Queenbee plugin from Python package.

    Usually you should not be using this function directly. Use ``load`` function
    instead.

    args:
        module: Python module object for a Queenbee Plugin module.

    returns:
        Plugin - A Queenbee plugin
    """
    qb_info = module.__pollination__
    package_name = module.__name__
    # get metadata
    config = PluginConfig.parse_obj(qb_info['config'])
    metadata = _get_meta_data(module, 'plugin')

    folder = pathlib.Path(module.__file__).parent

    functions = []
    for (_, name, _) in pkgutil.iter_modules([folder]):
        module = importlib.import_module('.' + name, package_name)
        for attr in dir(module):
            loaded_attr = getattr(module, attr)
            if hasattr(loaded_attr, '__decorator__') and \
                    getattr(loaded_attr, '__decorator__') == 'function':
                if loaded_attr is Function:
                    continue
                functions.append(loaded_attr().queenbee)

    plugin = Plugin(config=config, metadata=metadata, functions=functions)
    return plugin


def _load_recipe(module, baked: bool = False) -> Union[BakedRecipe, Recipe]:
    # load entry-point DAG
    """Load Queenbee plugin from Python package.

    Usually you should not be using this function directly. Use ``load`` function
    instead.

    args:
        module: Python module object for a Queenbee Recipe.

    returns:
        Recipe - A Queenbee recipe. It will be a baked recipe if baked is set to True.
    """
    qb_info = module.__pollination__
    package_name = module.__name__

    main_dag = qb_info.get('entry_point', None)()
    assert main_dag, f'{package_name} __pollination__ info is missing the enetry_point key.'

    # get metadata
    metadata = _get_meta_data(module, 'recipe')

    _dependencies = main_dag._dependencies
    # create a queenbee Recipe object

    # load dags
    qb_dag = main_dag.queenbee
    qb_dag.name = 'main'
    dags = [qb_dag] + [dag.queenbee for dag in _dependencies['dag']]

    # add dependencies
    repo = _init_repo()
    plugins = [
        Dependency(
            kind=DependencyKind.plugin, name=plugin['name'], tag=plugin['tag'],
            source=repo.as_uri()
        ) for plugin in _dependencies['plugin']
    ]
    recipes = [
        Dependency(
            kind=DependencyKind.recipe, name=recipe['name'], tag=recipe['tag'],
            source=repo.as_uri()
        ) for recipe in _dependencies['recipe']
    ]

    recipe = Recipe(metadata=metadata, dependencies=plugins + recipes, flow=dags)

    if baked:
        rf = RepositoryReference(name='pollination-dsl', path='file:///' + repo.as_posix())
        config = Config(repositories=[rf])
        recipe = BakedRecipe.from_recipe(recipe=recipe, config=config)

    return recipe


def load(package_name: str, baked: bool = False) -> Union[Plugin, BakedRecipe, Recipe]:
    """Load Queenbee Plugin or Recipe from Python package.

        package_name: Python package name (e.g. honeybee-radiance-pollination)
        baked: A boolean value to indicate wether to return a Recipe or a BakedRecipe.
    """
    module = import_module(package_name)

    assert hasattr(module, '__pollination__'), \
        'Failed to find __pollination__ info in __init__.py'
    qb_info = getattr(module, '__pollination__')
    if 'config' in qb_info:
        # it's a plugin
        return _load_plugin(module)
    else:
        # it's a recipe
        return _load_recipe(module, baked)


def package(package_name: str, readme: str = None) -> None:
    """Package a plugin or a recipe and add it to Queenbee local repository.

        Args:
            package_name: Python package name (e.g. honeybee-radiance-pollination)
            readme: Readme contents as a string.
    """
    # init a Queenbee package
    repository_path = _init_repo()
    index_path = repository_path/'index.json'
    qb_obj = load(package_name, baked=False)

    if isinstance(qb_obj, Recipe):
        qb_type = 'recipe'
    elif isinstance(qb_obj, Plugin):
        qb_type = 'plugin'
    else:
        raise TypeError('Input package must be a Queenbee Plugin or a Queenbee Recipe.')

    # add to pollination-dsl repository
    try:
        plugin_version, file_object = PackageVersion.package_resource(
            qb_obj, readme=_get_package_readme(package_name)
        )
    except Exception as error:
        raise ValueError(f'Failed to package {package_name} {qb_type}\n {error}')

    file_path = repository_path/f'{qb_type}s'/plugin_version.url
    file_object.seek(0)
    file_path.write_bytes(file_object.read())

    # re-index the repository
    repo_index = RepositoryIndex.from_folder(repository_path)
    repo_index.to_json(index_path.as_posix(), indent=2)


def translate(
        package_name: str,
        target_folder: str,
        baked: bool = False,
        readme: str = None
        ) -> str:
    """Translate Python package to a Queenbee plugin or recipe.

    BakedRecipes are written to a single file.

    args:
        package_name: Python package name. The package must be installed
            in the environment that this command being executed.
        target_folder: Path to folder to write this plugin.
        baked: A boolean to write the Recipes as BackedRecipes. It will be ignored for
            plugins.
        readme: Readme contents as a string.

    returns:
        str -- path to the generated folder or file.
    """
    qb_object = load(package_name, baked=baked)
    if baked and isinstance(qb_object, BakedRecipe):
        recipe_file = pathlib.Path(target_folder, package_name + '.yaml')
        recipe_file.write_text(qb_object.yaml())
        return recipe_file.as_posix()
    else:
        qb_object.to_folder(
            folder_path=target_folder,
            readme_string=_get_package_readme(package_name)
        )
        return target_folder



from typing import Dict
from dataclasses import dataclass
from queenbee.io.outputs.function import (
    FunctionStringOutput, FunctionIntegerOutput, FunctionNumberOutput,
    FunctionBooleanOutput, FunctionFolderOutput, FunctionFileOutput,
    FunctionPathOutput, FunctionJSONObjectOutput, FunctionArrayOutput
)
from queenbee.io.common import ItemType
from queenbee.base.basemodel import BaseModel
from queenbee.base.parser import parse_double_quotes_vars

from pydantic import validator, Field


__all__ = ('Outputs', )

_outputs_mapper = {
    'StringOutput': FunctionStringOutput,
    'IntegerOutput': FunctionIntegerOutput,
    'NumberOutput': FunctionNumberOutput,
    'BooleanOutput': FunctionBooleanOutput,
    'FolderOutput': FunctionFolderOutput,
    'FileOutput': FunctionFileOutput,
    'PathOutput': FunctionPathOutput,
    'DictOutput': FunctionJSONObjectOutput,
    'ListOutput': FunctionArrayOutput
}


class _OutputBase(BaseModel):

    @property
    def __decorator__(self) -> str:
        """Queenbee decorator for outputs."""
        return 'output'

    def to_queenbee(self, name):
        """Convert this output to a Queenbee output."""
        func = _outputs_mapper[self.__class__.__name__]
        data = {
            'name': name.replace('_', '-'),
            'path': self.path,
            'description': self.description,
            'annotations': self.annotations
        }

        if hasattr(self, 'items_type'):
            data['items_type'] = self.items_type

        return func.parse_obj(data)

    @property
    def is_artifact(self):
        return False

    @property
    def reference_type(self):
        return 'TaskReference'


class StringOutput(_OutputBase):
    """ A Function string output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source file for this output.

    """
    path: str
    annotations: Dict = None
    description: str = None

    @validator('path')
    def change_self_to_inputs(cls, v):
        refs = parse_double_quotes_vars(v)
        for ref in refs:
            v = v.replace(
                ref, ref.replace('self.', 'inputs.').replace('_', '-')
            )
        return v


class IntegerOutput(StringOutput):
    """ A Function integer output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source file for this output.

    """
    ...


class NumberOutput(StringOutput):
    """ A Function number output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source file for this output.

    """
    ...


class BooleanOutput(StringOutput):
    """ A Function boolean output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source file for this output.

    """
    ...


class DictOutput(StringOutput):
    """ A Function dictionary output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source file for this output.

    """
    ...


class ListOutput(StringOutput):
    """ A Function list output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source file for this output.

    """
    items_type: ItemType = Field(
        ItemType.String,
        description='Type of items in this array. All the items in an array must be '
        'from the same type.'
    )


class FolderOutput(StringOutput):
    """ A Function folder output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source folder for this output.

    """
    @property
    def is_artifact(self):
        return True

    @property
    def reference_type(self):
        return 'TaskFolderReference'


class FileOutput(FolderOutput):
    """ A Function file output.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source file for this output.

    """
    @property
    def reference_type(self):
        return 'TaskFileReference'


class PathOutput(FolderOutput):
    """ A Function path output. A path can be a file or a folder.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        path: Path to the source file or folder for this output.

    """
    @property
    def reference_type(self):
        return 'TaskPathReference'


@dataclass
class Outputs:
    """Function outputs."""
    str = StringOutput
    int = IntegerOutput
    float = NumberOutput
    bool = BooleanOutput
    file = FileOutput
    folder = FolderOutput
    path = PathOutput
    dict = DictOutput
    list = ListOutput



from typing import Any, Dict, List
from dataclasses import dataclass
from queenbee.io.inputs.dag import (
    DAGGenericInput, DAGStringInput, DAGIntegerInput, DAGNumberInput,
    DAGBooleanInput, DAGFolderInput, DAGFileInput, DAGPathInput,
    DAGJSONObjectInput, DAGArrayInput
)
from queenbee.base.basemodel import BaseModel, Field, validator
from queenbee.io.common import ItemType

from ..alias.inputs import InputAliasTypes


__all__ = ('Inputs', )

_inputs_mapper = {
    'GenericInput': DAGGenericInput,
    'StringInput': DAGStringInput,
    'IntegerInput': DAGIntegerInput,
    'NumberInput': DAGNumberInput,
    'BooleanInput': DAGBooleanInput,
    'FolderInput': DAGFolderInput,
    'FileInput': DAGFileInput,
    'PathInput': DAGPathInput,
    'DictInput': DAGJSONObjectInput,
    'ListInput': DAGArrayInput
}


class _InputBase(BaseModel):

    annotations: Dict = None
    description: str = None
    default: Any = None
    spec: Dict = None
    alias: List[InputAliasTypes] = None

    @validator('alias', always=True)
    def empty_list_alias(cls, v):
        return v if v is not None else []

    @property
    def __decorator__(self) -> str:
        """Queenbee decorator for inputs."""
        return 'input'

    def to_queenbee(self, name):
        """Convert this input to a Queenbee input."""
        func = _inputs_mapper[self.__class__.__name__]
        data = {
            'required': True if self.default is None else False,
            'name': name.replace('_', '-'),
            'default': self.default,
            'description': self.description,
            'annotations': self.annotations,
            'spec': self.spec,
            'alias': [al.to_queenbee().dict() for al in self.alias]
        }

        if hasattr(self, 'extensions'):
            data['extensions'] = self.extensions

        if hasattr(self, 'items_type'):
            data['items_type'] = self.items_type

        return func.parse_obj(data)

    @property
    def is_artifact(self):
        return False

    @property
    def reference_type(self):
        return 'InputReference'


class GenericInput(_InputBase):
    """ A DAG generic input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    ...


class StringInput(GenericInput):
    """ A DAG string input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    default: str = None


class IntegerInput(StringInput):
    """ A DAG integer input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    default: int = None


class NumberInput(StringInput):
    """ A DAG number input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    default: float = None


class BooleanInput(StringInput):
    """ A DAG boolean input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    default: bool = None


class DictInput(StringInput):
    """ A DAG dictionary input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    default: Dict = None


class ListInput(StringInput):
    """ A DAG list input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        items_type: 'Type of items in list. All the items in an array must be from '
        'the same type.'
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    default: List = None

    items_type: ItemType = Field(
        ItemType.String,
        description='Type of items in an array. All the items in an array must be from '
        'the same type.'
    )


class FolderInput(StringInput):
    """ A DAG folder input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    @property
    def is_artifact(self):
        return True

    @property
    def reference_type(self):
        return 'InputFolderReference'


class FileInput(FolderInput):
    """ A DAG file input.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        extensions: An optional list of valid extensions for input file.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """
    extensions: List[str] = None

    @property
    def reference_type(self):
        return 'InputFileReference'


class PathInput(FileInput):
    """ A DAG path input. A path can be a file or a folder.

    Args:
        annotations: An optional annotation dictionary.
        description: Input description.
        default: Default value.
        extensions: An optional list of valid extensions for input file.
        spec: A JSONSchema specification to validate input values.
        alias: A list of aliases for this input in different platforms.

    """

    @property
    def reference_type(self):
        return 'InputPathReference'


@dataclass
class Inputs:
    """DAG inputs enumeration."""
    any = GenericInput
    str = StringInput
    int = IntegerInput
    float = NumberInput
    bool = BooleanInput
    file = FileInput
    folder = FolderInput
    path = PathInput
    dict = DictInput
    list = ListInput
